"""Commands for the CLI for operations that can be performed."""

import asyncio
from typing import TYPE_CHECKING

import asyncclick as click
from asyncclick.core import Context

from myskoda.cli.utils import mqtt_required

if TYPE_CHECKING:
    from myskoda import MySkoda

from myskoda.models.air_conditioning import (
    AirConditioningAtUnlock,
    AirConditioningWithoutExternalPower,
    HeaterSource,
    SeatHeating,
    TargetTemperature,
    WindowHeating,
)
from myskoda.models.auxiliary_heating import AuxiliaryConfig, AuxiliaryStartMode


@click.command()
@click.option("temperature", "--temperature", type=float, required=True)
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def start_air_conditioning(
    ctx: Context,
    temperature: float,
    timeout: float,  # noqa: ASYNC109
    vin: str,
) -> None:
    """Start the air conditioning with the provided target temperature in °C."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.start_air_conditioning(vin, temperature)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def stop_air_conditioning(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Stop the air conditioning."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.stop_air_conditioning(vin)


@click.command()
@click.option("temperature", "--temperature", type=float, required=False)
@click.option("duration", "--duration", type=int, required=False)
@click.option(
    "mode", "--mode", type=click.Choice([e.value for e in AuxiliaryStartMode]), required=False
)
@click.option(
    "source", "--source", type=click.Choice([e.value for e in HeaterSource]), required=False
)
@click.option("spin", "--spin", type=str, required=True)
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def start_auxiliary_heating(  # noqa: PLR0913
    ctx: Context,
    spin: str,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    temperature: float | None = None,
    duration: int | None = None,
    mode: AuxiliaryStartMode | None = None,
    source: HeaterSource | None = None,
) -> None:
    """Start the auxiliary heating with the provided target temperature in °C."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    config = AuxiliaryConfig(
        target_temperature=TargetTemperature(temperature_value=temperature)
        if temperature is not None
        else None,
        duration_in_seconds=duration,
        start_mode=AuxiliaryStartMode(mode) if mode is not None else None,
        heater_source=HeaterSource(source) if source is not None else None,
    )
    async with asyncio.timeout(timeout):
        await myskoda.start_auxiliary_heating(vin, spin, config)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def stop_auxiliary_heating(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Stop the auxiliary heating."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.stop_auxiliary_heating(vin)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("temperature", "--temperature", type=float, required=True)
@click.pass_context
@mqtt_required
async def set_target_temperature(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    temperature: float,
) -> None:
    """Set the air conditioning's target temperature in °C."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_target_temperature(vin, temperature)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def start_window_heating(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Start heating both the front and rear window."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.start_window_heating(vin)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def stop_window_heating(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Stop heating both the front and rear window."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.stop_window_heating(vin)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("limit", "--limit", type=float, required=True)
@click.pass_context
@mqtt_required
async def set_charge_limit(ctx: Context, timeout: float, vin: str, limit: int) -> None:  # noqa: ASYNC109
    """Set the maximum charge limit in percent."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_charge_limit(vin, limit)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("limit", "--limit", type=float, required=True)
@click.pass_context
@mqtt_required
async def set_minimum_charge_limit(ctx: Context, timeout: float, vin: str, limit: int) -> None:  # noqa: ASYNC109
    """Set the minimum charge limit in percent."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_minimum_charge_limit(vin, limit)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_battery_care_mode(ctx: Context, timeout: float, vin: str, enabled: bool) -> None:  # noqa: ASYNC109
    """Enable or disable the battery care mode."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_battery_care_mode(vin, enabled)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_reduced_current_limit(ctx: Context, timeout: float, vin: str, enabled: bool) -> None:  # noqa: ASYNC109
    """Enable reducing the current limit by which the car is charged."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_reduced_current_limit(vin, enabled)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def wakeup(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Wake the vehicle up. Can be called maximum three times a day."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.wakeup(vin)


@click.command()
@click.option("spin", "--spin", type=str, required=True)
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def lock(
    ctx: Context,
    spin: str,
    timeout: float,  # noqa: ASYNC109
    vin: str,
) -> None:
    """Lock the car."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.lock(vin, spin)


@click.command()
@click.option("spin", "--spin", type=str, required=True)
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def unlock(
    ctx: Context,
    spin: str,
    timeout: float,  # noqa: ASYNC109
    vin: str,
) -> None:
    """Unlock the car."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.unlock(vin, spin)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def honk_flash(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Honk and/or flash."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.honk_flash(vin)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.pass_context
@mqtt_required
async def flash(ctx: Context, timeout: float, vin: str) -> None:  # noqa: ASYNC109
    """Flash."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.flash(vin)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_ac_without_external_power(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    enabled: bool,
) -> None:
    """Enable or disable AC without external power."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_ac_without_external_power(
            vin,
            AirConditioningWithoutExternalPower(
                air_conditioning_without_external_power_enabled=enabled
            ),
        )


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_ac_at_unlock(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    enabled: bool,
) -> None:
    """Enable or disable AC at unlock."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_ac_at_unlock(
            vin, AirConditioningAtUnlock(air_conditioning_at_unlock_enabled=enabled)
        )


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_windows_heating(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    enabled: bool,
) -> None:
    """Enable or disable windows heating with AC."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_windows_heating(vin, WindowHeating(window_heating_enabled=enabled))


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("front_left", "--frontLeft", type=bool, required=False)
@click.option("front_right", "--frontRight", type=bool, required=False)
@click.pass_context
@mqtt_required
async def set_seats_heating(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    front_left: bool | None = None,
    front_right: bool | None = None,
) -> None:
    """Enable or disable seats heating with AC."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    settings = SeatHeating(
        front_left=front_left,
        front_right=front_right,
    )
    async with asyncio.timeout(timeout):
        await myskoda.set_seats_heating(vin, settings)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_auto_unlock_plug(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    enabled: bool,
) -> None:
    """Enable or disable auto unlock plug when charged."""
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        await myskoda.set_auto_unlock_plug(vin, enabled)


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("timer", "--timer", type=click.Choice(["1", "2", "3"]), required=True)
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_departure_timer(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    timer: str,
    enabled: bool,
) -> None:
    """Enable or disable selected departure timer."""
    timer_id = int(timer)
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        # Get all timers from vehicle first
        departure_info = await myskoda.get_departure_timers(vin)
        if departure_info is not None:
            selected_timer = (
                next((t for t in departure_info.timers if t.id == timer_id), None)
                if departure_info.timers
                else None
            )
            if selected_timer is not None:
                selected_timer.enabled = enabled
                await myskoda.set_departure_timer(vin, selected_timer)
            else:
                print(f"No timer found with ID {timer_id}.")
        else:
            print("No DepartureInfo found for the given VIN.")


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("timer", "--timer", type=click.Choice(["1", "2", "3"]), required=True)
@click.option("enabled", "--enabled", type=bool, required=True)
@click.pass_context
@mqtt_required
async def set_ac_timer(
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    timer: str,
    enabled: bool,
) -> None:
    """Enable or disable selected air-conditioning timer."""
    timer_id = int(timer)
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        # Get all timers from vehicle first
        air_conditioning = await myskoda.get_air_conditioning(vin)
        if air_conditioning is not None:
            selected_timer = (
                next((t for t in air_conditioning.timers if t.id == timer_id), None)
                if air_conditioning.timers
                else None
            )
            if selected_timer is not None:
                selected_timer.enabled = enabled
                await myskoda.set_ac_timer(vin, selected_timer)
            else:
                print(f"No timer found with ID {timer_id}.")
        else:
            print("No AirConditioning found for the given VIN.")


@click.command()
@click.option("timeout", "--timeout", type=float, default=300)
@click.argument("vin")
@click.option("timer", "--timer", type=click.Choice(["1", "2", "3"]), required=True)
@click.option("enabled", "--enabled", type=bool, required=True)
@click.option("spin", "--spin", type=str, required=True)
@click.pass_context
@mqtt_required
async def set_aux_timer(  # noqa: PLR0913
    ctx: Context,
    timeout: float,  # noqa: ASYNC109
    vin: str,
    timer: str,
    enabled: bool,
    spin: str,
) -> None:
    """Enable or disable selected auxiliary-heating timer."""
    timer_id = int(timer)
    myskoda: MySkoda = ctx.obj["myskoda"]
    async with asyncio.timeout(timeout):
        # Get all timers from vehicle first
        auxiliary_heating = await myskoda.get_auxiliary_heating(vin)
        if auxiliary_heating is not None:
            selected_timer = (
                next((t for t in auxiliary_heating.timers if t.id == timer_id), None)
                if auxiliary_heating.timers
                else None
            )
            if selected_timer is not None:
                selected_timer.enabled = enabled
                await myskoda.set_auxiliary_heating_timer(vin, selected_timer, spin)
            else:
                print(f"No timer found with ID {timer_id}.")
        else:
            print("No AuxiliaryHeating found for the given VIN.")



"""Main entry point for the MySkoda library.

This class provides all methods to operate on the API and MQTT broker.
"""

import logging
from asyncio import gather, timeout
from collections.abc import Awaitable, Callable
from copy import deepcopy
from datetime import UTC, datetime
from ssl import SSLContext
from traceback import format_exc
from types import SimpleNamespace
from typing import Any

from aiohttp import ClientSession, TraceConfig, TraceRequestEndParams

from myskoda.anonymize import anonymize_url
from myskoda.models.fixtures import (
    Endpoint,
    Fixture,
    FixtureReportGet,
    FixtureReportType,
    FixtureVehicle,
    create_fixture_vehicle,
)

from .__version__ import __version__ as version
from .auth.authorization import Authorization
from .const import MQTT_OPERATION_TIMEOUT
from .event import Event
from .models.air_conditioning import (
    AirConditioning,
    AirConditioningAtUnlock,
    AirConditioningTimer,
    AirConditioningWithoutExternalPower,
    SeatHeating,
    WindowHeating,
)
from .models.auxiliary_heating import AuxiliaryConfig, AuxiliaryHeating, AuxiliaryHeatingTimer
from .models.charging import ChargeMode, Charging
from .models.departure import DepartureInfo, DepartureTimer
from .models.driving_range import DrivingRange
from .models.health import Health
from .models.info import CapabilityId, Info
from .models.maintenance import Maintenance
from .models.operation_request import OperationName
from .models.position import Positions
from .models.spin import Spin
from .models.status import Status
from .models.trip_statistics import TripStatistics
from .models.user import User
from .mqtt import MySkodaMqttClient
from .rest_api import GetEndpointResult, RestApi
from .vehicle import Vehicle

_LOGGER = logging.getLogger(__name__)

type Vin = str


async def trace_response(
    _session: ClientSession,
    _trace_config_ctx: SimpleNamespace,
    params: TraceRequestEndParams,
) -> None:
    """Log response details. Used in aiohttp.TraceConfig."""
    resp_text = await params.response.text()
    _LOGGER.debug(
        "Trace: %s %s - response: %s (%s bytes) %s",
        params.method,
        str(params.url)[:60],
        params.response.status,
        params.response.content_length,
        resp_text[:5000],
    )


TRACE_CONFIG = TraceConfig()
TRACE_CONFIG.on_request_end.append(trace_response)


class MySkoda:
    session: ClientSession
    rest_api: RestApi
    mqtt: MySkodaMqttClient | None = None
    authorization: Authorization
    ssl_context: SSLContext | None = None
    vehicles: dict[Vin, Vehicle]

    def __init__(  # noqa: D107, PLR0913
        self,
        session: ClientSession,
        ssl_context: SSLContext | None = None,
        mqtt_enabled: bool = True,
        mqtt_broker_host: str | None = None,
        mqtt_broker_port: int | None = None,
        mqtt_enable_ssl: bool | None = None,
    ) -> None:
        self.vehicles: dict[Vin, Vehicle] = {}
        self.session = session
        self.authorization = Authorization(session)
        self.rest_api = RestApi(self.session, self.authorization)
        self.ssl_context = ssl_context
        self.mqtt_broker_host = mqtt_broker_host
        self.mqtt_broker_port = mqtt_broker_port
        self.mqtt_enable_ssl = mqtt_enable_ssl
        if mqtt_enabled:
            self.mqtt = self._create_mqtt_client()

    def _create_mqtt_client(self) -> MySkodaMqttClient:
        kwargs = {
            "authorization": self.authorization,
            "hostname": self.mqtt_broker_host,
            "port": self.mqtt_broker_port,
            "enable_ssl": self.mqtt_enable_ssl,
        }
        return MySkodaMqttClient(**{k: v for k, v in kwargs.items() if v is not None})

    async def enable_mqtt(self) -> None:
        """If MQTT was not enabled when initializing MySkoda, enable it manually and connect."""
        if self.mqtt is not None:
            return
        self.mqtt = self._create_mqtt_client()
        user = await self.get_user()
        vehicles = await self.list_vehicle_vins()
        await self.mqtt.connect(user.id, vehicles)

    async def _wait_for_operation(self, operation: OperationName) -> None:
        if self.mqtt is None:
            return
        try:
            async with timeout(MQTT_OPERATION_TIMEOUT):
                await self.mqtt.wait_for_operation(operation)
        except TimeoutError:
            _LOGGER.warning("Timeout occurred while waiting for %s. Aborted.", operation)

    async def connect(self, email: str, password: str) -> None:
        """Authenticate on the rest api and connect to the MQTT broker."""
        await self.authorization.authorize(email, password)
        _LOGGER.debug("IDK Authorization was successful.")

        if self.mqtt:
            user = await self.get_user()
            vehicles = await self.list_vehicle_vins()
            await self.mqtt.connect(user.id, vehicles)
        _LOGGER.debug("MySkoda ready.")

    def subscribe(self, callback: Callable[[Event], None | Awaitable[None]]) -> None:
        """Listen for events emitted by MySkoda's MQTT broker."""
        if self.mqtt is None:
            raise MqttDisabledError
        self.mqtt.subscribe(callback)

    async def disconnect(self) -> None:
        """Disconnect from the MQTT broker."""
        if self.mqtt:
            await self.mqtt.disconnect()

    async def stop_charging(self, vin: str) -> None:
        """Stop the car from charging."""
        future = self._wait_for_operation(OperationName.STOP_CHARGING)
        await self.rest_api.stop_charging(vin)
        await future

    async def start_charging(self, vin: str) -> None:
        """Start charging the car."""
        future = self._wait_for_operation(OperationName.START_CHARGING)
        await self.rest_api.start_charging(vin)
        await future

    async def set_charge_mode(self, vin: str, mode: ChargeMode) -> None:
        """Set the charge mode."""
        future = self._wait_for_operation(OperationName.UPDATE_CHARGE_MODE)
        await self.rest_api.set_charge_mode(vin, mode=mode)
        await future

    async def honk_flash(self, vin: str) -> None:
        """Honk and flash."""
        future = self._wait_for_operation(OperationName.START_HONK)
        await self.rest_api.honk_flash(vin, (await self.get_positions(vin)).positions)
        await future

    async def flash(self, vin: str) -> None:
        """Flash lights."""
        future = self._wait_for_operation(OperationName.START_FLASH)
        await self.rest_api.flash(vin, (await self.get_positions(vin)).positions)
        await future

    async def wakeup(self, vin: str) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        future = self._wait_for_operation(OperationName.WAKEUP)
        await self.rest_api.wakeup(vin)
        await future

    async def set_reduced_current_limit(self, vin: str, reduced: bool) -> None:
        """Enable reducing the current limit by which the car is charged."""
        future = self._wait_for_operation(OperationName.UPDATE_CHARGING_CURRENT)
        await self.rest_api.set_reduced_current_limit(vin, reduced=reduced)
        await future

    async def set_battery_care_mode(self, vin: str, enabled: bool) -> None:
        """Enable or disable the battery care mode."""
        future = self._wait_for_operation(OperationName.UPDATE_CARE_MODE)
        await self.rest_api.set_battery_care_mode(vin, enabled)
        await future

    async def set_auto_unlock_plug(self, vin: str, enabled: bool) -> None:
        """Enable or disable auto unlock plug when charged."""
        future = self._wait_for_operation(OperationName.UPDATE_AUTO_UNLOCK_PLUG)
        await self.rest_api.set_auto_unlock_plug(vin, enabled)
        await future

    async def set_charge_limit(self, vin: str, limit: int) -> None:
        """Set the maximum charge limit in percent."""
        future = self._wait_for_operation(OperationName.UPDATE_CHARGE_LIMIT)
        await self.rest_api.set_charge_limit(vin, limit)
        await future

    async def set_minimum_charge_limit(self, vin: str, limit: int) -> None:
        """Set minimum battery SoC in percent for departure timer."""
        future = self._wait_for_operation(OperationName.UPDATE_MINIMAL_SOC)
        await self.rest_api.set_minimum_charge_limit(vin, limit)
        await future

    async def stop_window_heating(self, vin: str) -> None:
        """Stop heating both the front and rear window."""
        future = self._wait_for_operation(OperationName.STOP_WINDOW_HEATING)
        await self.rest_api.stop_window_heating(vin)
        await future

    async def start_window_heating(self, vin: str) -> None:
        """Start heating both the front and rear window."""
        future = self._wait_for_operation(OperationName.START_WINDOW_HEATING)
        await self.rest_api.start_window_heating(vin)
        await future

    async def set_ac_without_external_power(
        self, vin: str, settings: AirConditioningWithoutExternalPower
    ) -> None:
        """Enable or disable AC without external power."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_WITHOUT_EXTERNAL_POWER)
        await self.rest_api.set_ac_without_external_power(vin, settings)
        await future

    async def set_ac_at_unlock(self, vin: str, settings: AirConditioningAtUnlock) -> None:
        """Enable or disable AC at unlock."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_AT_UNLOCK)
        await self.rest_api.set_ac_at_unlock(vin, settings)
        await future

    async def set_windows_heating(self, vin: str, settings: WindowHeating) -> None:
        """Enable or disable windows heating with AC."""
        future = self._wait_for_operation(OperationName.WINDOWS_HEATING)
        await self.rest_api.set_windows_heating(vin, settings)
        await future

    async def set_seats_heating(self, vin: str, settings: SeatHeating) -> None:
        """Enable or disable seats heating with AC."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_SEATS_HEATING)
        await self.rest_api.set_seats_heating(vin, settings)
        await future

    async def set_target_temperature(self, vin: str, temperature: float) -> None:
        """Set the air conditioning's target temperature in °C."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_TARGET_TEMPERATURE)
        await self.rest_api.set_target_temperature(vin, temperature)
        await future

    async def start_air_conditioning(self, vin: str, temperature: float) -> None:
        """Start the air conditioning with the provided target temperature in °C."""
        future = self._wait_for_operation(OperationName.START_AIR_CONDITIONING)
        await self.rest_api.start_air_conditioning(vin, temperature)
        await future

    async def stop_air_conditioning(self, vin: str) -> None:
        """Stop the air conditioning."""
        future = self._wait_for_operation(OperationName.STOP_AIR_CONDITIONING)
        await self.rest_api.stop_air_conditioning(vin)
        await future

    async def start_auxiliary_heating(
        self, vin: str, spin: str, config: AuxiliaryConfig | None = None
    ) -> None:
        """Start the auxiliary heating with the provided configuration."""
        future = self._wait_for_operation(OperationName.START_AUXILIARY_HEATING)
        await self.rest_api.start_auxiliary_heating(vin, spin, config=config)
        await future

    async def stop_auxiliary_heating(self, vin: str) -> None:
        """Stop the auxiliary heating."""
        future = self._wait_for_operation(OperationName.STOP_AUXILIARY_HEATING)
        await self.rest_api.stop_auxiliary_heating(vin)
        await future

    async def set_ac_timer(self, vin: str, timer: AirConditioningTimer) -> None:
        """Send provided air-conditioning timer to the vehicle."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_TIMERS)
        await self.rest_api.set_ac_timer(vin, timer)
        await future

    async def set_auxiliary_heating_timer(
        self, vin: str, timer: AuxiliaryHeatingTimer, spin: str
    ) -> None:
        """Send provided auxiliary heating timer to the vehicle."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_TIMERS)
        await self.rest_api.set_auxiliary_heating_timer(vin, timer, spin)
        await future

    async def lock(self, vin: str, spin: str) -> None:
        """Lock the car."""
        future = self._wait_for_operation(OperationName.LOCK)
        await self.rest_api.lock(vin, spin)
        await future

    async def unlock(self, vin: str, spin: str) -> None:
        """Unlock the car."""
        future = self._wait_for_operation(OperationName.UNLOCK)
        await self.rest_api.unlock(vin, spin)
        await future

    async def set_departure_timer(self, vin: str, timer: DepartureTimer) -> None:
        """Send provided departure timer to the vehicle."""
        future = self._wait_for_operation(OperationName.UPDATE_DEPARTURE_TIMERS)
        await self.rest_api.set_departure_timer(vin, timer)
        await future

    async def get_departure_timers(self, vin: str, anonymize: bool = False) -> DepartureInfo:
        """Retrieve departure timers for the specified vehicle."""
        return (await self.rest_api.get_departure_timers(vin, anonymize=anonymize)).result

    async def get_auth_token(self) -> str:
        """Retrieve the main access token for the IDK session."""
        return await self.rest_api.authorization.get_access_token()

    async def verify_spin(self, spin: str, anonymize: bool = False) -> Spin:
        """Verify S-PIN."""
        return (await self.rest_api.verify_spin(spin, anonymize=anonymize)).result

    async def get_info(self, vin: str, anonymize: bool = False) -> Info:
        """Retrieve the basic vehicle information for the specified vehicle."""
        return (await self.rest_api.get_info(vin, anonymize=anonymize)).result

    async def get_charging(self, vin: str, anonymize: bool = False) -> Charging:
        """Retrieve information related to charging for the specified vehicle."""
        return (await self.rest_api.get_charging(vin, anonymize=anonymize)).result

    async def get_status(self, vin: str, anonymize: bool = False) -> Status:
        """Retrieve the current status for the specified vehicle."""
        return (await self.rest_api.get_status(vin, anonymize=anonymize)).result

    async def get_air_conditioning(self, vin: str, anonymize: bool = False) -> AirConditioning:
        """Retrieve the current air conditioning status for the specified vehicle."""
        return (await self.rest_api.get_air_conditioning(vin, anonymize=anonymize)).result

    async def get_auxiliary_heating(self, vin: str, anonymize: bool = False) -> AuxiliaryHeating:
        """Retrieve the current auxiliary heating status for the specified vehicle."""
        return (await self.rest_api.get_auxiliary_heating(vin, anonymize=anonymize)).result

    async def get_positions(self, vin: str, anonymize: bool = False) -> Positions:
        """Retrieve the current position for the specified vehicle."""
        return (await self.rest_api.get_positions(vin, anonymize=anonymize)).result

    async def get_driving_range(self, vin: str, anonymize: bool = False) -> DrivingRange:
        """Retrieve estimated driving range for combustion vehicles."""
        return (await self.rest_api.get_driving_range(vin, anonymize=anonymize)).result

    async def get_trip_statistics(self, vin: str, anonymize: bool = False) -> TripStatistics:
        """Retrieve statistics about past trips."""
        return (await self.rest_api.get_trip_statistics(vin, anonymize=anonymize)).result

    async def get_maintenance(self, vin: str, anonymize: bool = False) -> Maintenance:
        """Retrieve maintenance report."""
        return (await self.rest_api.get_maintenance(vin, anonymize=anonymize)).result

    async def get_health(self, vin: str, anonymize: bool = False) -> Health:
        """Retrieve health information for the specified vehicle."""
        return (await self.rest_api.get_health(vin, anonymize=anonymize)).result

    async def get_user(self, anonymize: bool = False) -> User:
        """Retrieve user information about logged in user."""
        return (await self.rest_api.get_user(anonymize=anonymize)).result

    async def list_vehicle_vins(self) -> list[str]:
        """List all vehicles by their vins."""
        garage = (await self.rest_api.get_garage()).result
        if garage.vehicles is None:
            return []
        return [vehicle.vin for vehicle in garage.vehicles]

    async def get_vehicle(
        self, vin: str, excluded_capabilities: list[CapabilityId] | None = None
    ) -> Vehicle:
        """Load a full vehicle based on its capabilities."""
        capabilities = [
            CapabilityId.AIR_CONDITIONING,
            CapabilityId.AUXILIARY_HEATING,
            CapabilityId.CHARGING,
            CapabilityId.PARKING_POSITION,
            CapabilityId.STATE,
            CapabilityId.TRIP_STATISTICS,
            CapabilityId.VEHICLE_HEALTH_INSPECTION,
            CapabilityId.DEPARTURE_TIMERS,
        ]

        if excluded_capabilities:
            capabilities = [c for c in capabilities if c not in excluded_capabilities]

        return await self.get_partial_vehicle(vin, capabilities)

    async def get_partial_vehicle(self, vin: str, capabilities: list[CapabilityId]) -> Vehicle:
        """Load a partial vehicle, based on list of capabilities."""
        info = await self.get_info(vin)
        maintenance = await self.get_maintenance(vin)

        if vin in self.vehicles:
            self.vehicles[vin].info = info
            self.vehicles[vin].maintenance = maintenance
        else:
            self.vehicles[vin] = Vehicle(info, maintenance)

        for capa in capabilities:
            if info.is_capability_available(capa):
                await self._request_capability_data(vin, capa)

        return deepcopy(self.vehicles[vin])

    async def _request_capability_data(self, vin: str, capa: CapabilityId) -> None:
        """Request specific capability data from MySkoda API."""
        try:
            match capa:
                case CapabilityId.AIR_CONDITIONING:
                    self.vehicles[vin].air_conditioning = await self.get_air_conditioning(vin)
                case CapabilityId.AUXILIARY_HEATING:
                    self.vehicles[vin].auxiliary_heating = await self.get_auxiliary_heating(vin)
                case CapabilityId.CHARGING:
                    self.vehicles[vin].charging = await self.get_charging(vin)
                case CapabilityId.PARKING_POSITION:
                    self.vehicles[vin].positions = await self.get_positions(vin)
                case CapabilityId.STATE:
                    self.vehicles[vin].status = await self.get_status(vin)
                    self.vehicles[vin].driving_range = await self.get_driving_range(vin)
                case CapabilityId.TRIP_STATISTICS:
                    self.vehicles[vin].trip_statistics = await self.get_trip_statistics(vin)
                case CapabilityId.VEHICLE_HEALTH_INSPECTION:
                    self.vehicles[vin].health = await self.get_health(vin)
                case CapabilityId.DEPARTURE_TIMERS:
                    self.vehicles[vin].departure_info = await self.get_departure_timers(vin)
        except Exception as err:  # noqa: BLE001
            _LOGGER.warning("Requesting %s failed: %s, continue", capa, err)

    async def get_all_vehicles(self) -> list[Vehicle]:
        """Load all vehicles based on their capabilities."""
        vins = await self.list_vehicle_vins()
        return await gather(*(self.get_vehicle(vin) for vin in vins))

    async def generate_fixture_report(
        self, vin: str, vehicle: FixtureVehicle, endpoint: Endpoint
    ) -> FixtureReportGet:
        """Generate a fixture report for the specified endpoint and vehicle."""
        try:
            result = await self.get_endpoint(vin, endpoint, anonymize=True)
        except Exception:  # noqa: BLE001
            return FixtureReportGet(
                type=FixtureReportType.GET,
                vehicle_id=vehicle.id,
                success=False,
                endpoint=endpoint,
                error=anonymize_url(format_exc()),
            )
        else:
            return FixtureReportGet(
                type=FixtureReportType.GET,
                vehicle_id=vehicle.id,
                raw=result.raw,
                success=True,
                url=result.url,
                endpoint=endpoint,
                result=result.result.to_dict(),
            )

    async def get_endpoint(
        self, vin: str, endpoint: Endpoint, anonymize: bool = False
    ) -> GetEndpointResult[Any]:
        """Invoke a get endpoint by endpoint enum."""
        # Mapping of endpoints to corresponding methods
        endpoint_method_map = {
            Endpoint.INFO: self.rest_api.get_info,
            Endpoint.STATUS: self.rest_api.get_status,
            Endpoint.AIR_CONDITIONING: self.rest_api.get_air_conditioning,
            Endpoint.AUXILIARY_HEATING: self.rest_api.get_auxiliary_heating,
            Endpoint.POSITIONS: self.rest_api.get_positions,
            Endpoint.HEALTH: self.rest_api.get_health,
            Endpoint.CHARGING: self.rest_api.get_charging,
            Endpoint.MAINTENANCE: self.rest_api.get_maintenance,
            Endpoint.DRIVING_RANGE: self.rest_api.get_driving_range,
            Endpoint.TRIP_STATISTICS: self.rest_api.get_trip_statistics,
            Endpoint.DEPARTURE_INFO: self.rest_api.get_departure_timers,
        }

        # Look up the method, or raise an error if unsupported
        method = endpoint_method_map.get(endpoint)
        if not method:
            error_message = f"Unsupported endpoint: {endpoint}"
            raise UnsupportedEndpointError(error_message)

        # Call the method and return the result
        return await method(vin, anonymize=anonymize)

    async def generate_get_fixture(
        self, name: str, description: str, vins: list[str], endpoint: Endpoint
    ) -> Fixture:
        """Generate a fixture for a get request."""
        vehicles = [
            (vin, create_fixture_vehicle(i, await self.get_info(vin))) for i, vin in enumerate(vins)
        ]

        endpoints = []
        if endpoint != Endpoint.ALL:
            endpoints = [endpoint]
        else:
            endpoints = filter(lambda ep: ep != Endpoint.ALL, Endpoint)

        reports = [
            await self.generate_fixture_report(vin, vehicle, endpoint)
            for (vin, vehicle) in vehicles
            for endpoint in endpoints
        ]

        return Fixture(
            name=name,
            description=description,
            generation_time=datetime.now(tz=UTC),
            vehicles=[vehicle for (_, vehicle) in vehicles],
            reports=reports,
            library_version=version,
        )


class MqttDisabledError(Exception):
    """MQTT was not enabled."""


class UnsupportedEndpointError(Exception):
    """Endpoint not implemented."""



"""Methods for anonymizing data from the API."""

import re

ACCESS_TOKEN = "eyJ0eXAiOiI0ODEyODgzZi05Y2FiLTQwMWMtYTI5OC0wZmEyMTA5Y2ViY2EiLCJhbGciOiJSUzI1NiJ9"  # noqa: S105
USER_ID = "b8bc126c-ee36-402b-8723-2c1c3dff8dec"
VIN = "TMOCKAA0AA000000"
VIN_REGEX = re.compile(r"TMB\w{14}")
ADDRESS = {
    "city": "Example City",
    "street": "Example Avenue",
    "houseNumber": "15",
    "zipCode": "54321",
    "countryCode": "DEU",
}
SERVICE_PARTNER_ID = "DEU11111"
PARTNER_NUMBER = "1111"
PARTNER_NAME = "Example Service Partner"
LOCATION = {
    "latitude": 53.470636,
    "longitude": 9.689872,
}
EMAIL = "user@example.com"
PHONE = "+49 1234 567890"
VEHICLE_NAME = "Example Car"
LICENSE_PLATE = "HH AA 1234"
URL = "https://example.com"
FIRST_NAME = "John"
LAST_NAME = "Dough"
NICKNAME = "Johnny D."
PROFILE_PICTURE_URL = "https://example.com/profile.jpg"
DATE_OF_BIRTH = "2000-01-01"

SERVICE_PARTNER = {
    "name": PARTNER_NAME,
    "partnerNumber": PARTNER_NUMBER,
    "id": SERVICE_PARTNER_ID,
    "contact": {
        "phone": PHONE,
        "url": URL,
        "email": EMAIL,
    },
    "address": ADDRESS,
    "location": LOCATION,
}


def anonymize_info(data: dict) -> dict:
    data["vin"] = VIN
    data["name"] = VEHICLE_NAME
    if "licensePlate" in data:
        data["licensePlate"] = LICENSE_PLATE
    if "servicePartner" in data:
        data["servicePartner"]["servicePartnerId"] = SERVICE_PARTNER_ID
    return data


def anonymize_maintenance(data: dict) -> dict:
    if "preferredServicePartner" in data:
        data["preferredServicePartner"].update(SERVICE_PARTNER)
    if "predictiveMaintenance" in data:
        data["predictiveMaintenance"]["setting"]["email"] = EMAIL
        data["predictiveMaintenance"]["setting"]["phone"] = PHONE
    for booking in data.get("customerService", {}).get("bookingHistory", []):
        booking["servicePartner"].update(SERVICE_PARTNER)
    for booking in data.get("customerService", {}).get("bookingHistory", []):
        booking["servicePartner"].update(SERVICE_PARTNER)
    return data


def anonymize_charging(data: dict) -> dict:
    return data


def anonymize_status(data: dict) -> dict:
    return data


def anonymize_air_conditioning(data: dict) -> dict:
    return data


def anonymize_auxiliary_heating(data: dict) -> dict:
    return data


def anonymize_departure_timers(data: dict) -> dict:
    return data


def anonymize_positions(data: dict) -> dict:
    if "positions" in data:
        for position in data["positions"]:
            position["gpsCoordinates"] = LOCATION
            position["address"] = ADDRESS
    return data


def anonymize_driving_range(data: dict) -> dict:
    return data


def anonymize_trip_statistics(data: dict) -> dict:
    return data


def anonymize_health(data: dict) -> dict:
    return data


def anonymize_user(data: dict) -> dict:
    data["email"] = EMAIL
    data["firstName"] = FIRST_NAME
    data["lastName"] = LAST_NAME
    data["nickname"] = NICKNAME
    data["profilePictureUrl"] = PROFILE_PICTURE_URL
    data["dateOfBirth"] = DATE_OF_BIRTH
    data["phone"] = PHONE

    return data


def anonymize_garage_entry(data: dict) -> dict:
    data["vin"] = VIN
    data["name"] = VEHICLE_NAME
    return data


def anonymize_garage(data: dict) -> dict:
    if "vehicles" in data:
        data["vehicles"] = [anonymize_garage_entry(vehicle) for vehicle in data["vehicles"]]
    return data


def anonymize_url(url: str) -> str:
    return VIN_REGEX.sub(VIN, url)



"""Common models used in multiple responses."""

from dataclasses import dataclass, field
from datetime import UTC, datetime
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin


class CaseInsensitiveStrEnum(StrEnum):
    @classmethod
    def _missing_(cls, value: object) -> StrEnum | None:
        """Ignore the case of the value.

        Some endpoints will return values sometimes as uppercase and sometimes as lowercase...
        """
        if not isinstance(value, str):
            raise TypeError
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        return None


class OnOffState(StrEnum):
    ON = "ON"
    OFF = "OFF"
    INVALID = "INVALID"


class EnabledState(StrEnum):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"
    NOT_ALLOWED = "NOT_ALLOWED"


class ActiveState(StrEnum):
    ACTIVATED = "ACTIVATED"
    DEACTIVATED = "DEACTIVATED"


class OpenState(StrEnum):
    OPEN = "OPEN"
    CLOSED = "CLOSED"
    UNSUPPORTED = "UNSUPPORTED"
    UNKNOWN = "UNKNOWN"


class DoorLockedState(StrEnum):
    LOCKED = "YES"
    OPENED = "OPENED"
    TRUNK_OPENED = "TRUNK_OPENED"
    UNLOCKED = "NO"
    UNKNOWN = "UNKNOWN"


class ChargerLockedState(StrEnum):
    LOCKED = "LOCKED"
    UNLOCKED = "UNLOCKED"
    INVALID = "INVALID"


class ConnectionState(StrEnum):
    CONNECTED = "CONNECTED"
    DISCONNECTED = "DISCONNECTED"


class Side(StrEnum):
    LEFT = "LEFT"
    RIGHT = "RIGHT"


@dataclass
class Coordinates(DataClassORJSONMixin):
    """GPS Coordinates."""

    latitude: float
    longitude: float


@dataclass
class Address(DataClassORJSONMixin):
    """A representation of a house-address."""

    country_code: str = field(metadata=field_options(alias="countryCode"))
    zip_code: str | None = field(default=None, metadata=field_options(alias="zipCode"))
    house_number: str | None = field(default=None, metadata=field_options(alias="houseNumber"))
    street: str | None = field(default=None)
    city: str | None = field(default=None)
    country: str | None = field(default=None)


class Weekday(StrEnum):
    MONDAY = "MONDAY"
    TUESDAY = "TUESDAY"
    WEDNESDAY = "WEDNESDAY"
    THURSDAY = "THURSDAY"
    FRIDAY = "FRIDAY"
    SATURDAY = "SATURDAY"
    SUNDAY = "SUNDAY"


@dataclass
class BaseResponse(DataClassORJSONMixin):
    """Base class for all API response models.

    All responses have the current timestamp injected.
    """

    timestamp: datetime = field(default_factory=lambda: datetime.now(UTC), kw_only=True)



"""Models related to vehicle events from the MQTT broker."""

from dataclasses import dataclass, field
from datetime import datetime
from enum import StrEnum
from typing import Generic, TypeVar

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from myskoda.models.vehicle_ignition_status import IgnitionStatus, UnexpectedIgnitionStatusError


class VehicleEventName(StrEnum):
    """List of known vehicle EventNames."""

    VEHICLE_AWAKE = "vehicle-awake"
    VEHICLE_CONNECTION_ONLINE = "vehicle-connection-online"
    VEHICLE_WARNING_BATTEYLEVEL = "vehicle-warning-batterylevel"
    VEHICLE_IGNITION_STATUS_CHANGED = "vehicle-ignition-status-changed"


@dataclass
class VehicleEventData(DataClassORJSONMixin):
    """Base class for data in vehicle events."""

    user_id: str = field(metadata=field_options(alias="userId"))
    vin: str


T = TypeVar("T", bound=VehicleEventData)


@dataclass
class VehicleEvent(Generic[T], DataClassORJSONMixin):
    """Main model for Vehicle Events.

    Vehicle Events are unsolicited events emitted by the MQTT bus towards the client.
    """

    version: int
    producer: str
    name: VehicleEventName
    data: T
    trace_id: str = field(metadata=field_options(alias="traceId"))
    timestamp: datetime | None = field(default=None)


def _deserialize_ignition_status(value: str) -> IgnitionStatus:
    match value:
        case "ON":
            return IgnitionStatus.ON
        case "OFF":
            return IgnitionStatus.OFF
        case _:
            raise UnexpectedIgnitionStatusError


@dataclass
class VehicleEventVehicleIgnitionStatusData(VehicleEventData):
    """Ignition data inside vehicle service event vehicle-ignition-status."""

    ignition_status: IgnitionStatus | None = field(
        default=None,
        metadata=field_options(alias="ignitionStatus", deserialize=_deserialize_ignition_status),
    )


@dataclass
class VehicleEventWithVehicleIgnitionStatusData(VehicleEvent):
    data: VehicleEventVehicleIgnitionStatusData



"""MQTT client module for the MySkoda server.

Inspired by https://github.com/YoSmart-Inc/yolink-api/tree/main
"""

import asyncio
import logging
import re
import ssl
import uuid
from collections.abc import Awaitable, Callable
from datetime import UTC, datetime
from random import uniform
from typing import Any, cast

import aiomqtt

from myskoda.auth.authorization import Authorization
from myskoda.models.service_event import ServiceEvent, ServiceEventWithChargingData
from myskoda.models.vehicle_event import VehicleEvent, VehicleEventWithVehicleIgnitionStatusData

from .const import (
    MQTT_ACCOUNT_EVENT_TOPICS,
    MQTT_BROKER_HOST,
    MQTT_BROKER_PORT,
    MQTT_FAST_RETRY,
    MQTT_KEEPALIVE,
    MQTT_MAX_RECONNECT_DELAY,
    MQTT_OPERATION_TOPICS,
    MQTT_RECONNECT_DELAY,
    MQTT_SERVICE_EVENT_TOPICS,
    MQTT_VEHICLE_EVENT_TOPICS,
)
from .event import (
    Event,
    EventAccess,
    EventAccountPrivacy,
    EventAirConditioning,
    EventAuxiliaryHeating,
    EventCharging,
    EventDeparture,
    EventLights,
    EventOdometer,
    EventOperation,
    EventType,
    EventVehicleConnectionStatusUpdate,
    EventVehicleIgnitionStatus,
)
from .models.operation_request import OperationName, OperationRequest, OperationStatus

_LOGGER = logging.getLogger(__name__)
TOPIC_RE = re.compile("^(.*?)/(.*?)/(.*?)/(.*?)$")
app_uuid = uuid.uuid4()


def _create_ssl_context() -> ssl.SSLContext:
    """Create a SSL context for the MQTT connection."""
    context = ssl.SSLContext(ssl.PROTOCOL_TLS)
    context.load_default_certs()
    return context


_SSL_CONTEXT = _create_ssl_context()

background_tasks = set()


class OperationListener:
    """Used to track callbacks to execute for a given OperationName."""

    operation_name: OperationName
    future: asyncio.Future[OperationRequest]

    def __init__(  # noqa: D107
        self, operation_name: OperationName, future: asyncio.Future[OperationRequest]
    ) -> None:
        self.operation_name = operation_name
        self.future = future


class OperationFailedError(Exception):
    def __init__(self, operation: OperationRequest) -> None:  # noqa: D107
        op = operation.operation
        error = operation.error_code
        trace = operation.trace_id
        super().__init__(f"Operation {op} with trace {trace} failed: {error}")


class MySkodaMqttClient:
    user_id: str | None
    vehicle_vins: list[str]
    _callbacks: list[Callable[[Event], None | Awaitable[None]]]
    _operation_listeners: list[OperationListener]

    def __init__(  # noqa: D107
        self,
        authorization: Authorization,
        hostname: str = MQTT_BROKER_HOST,
        port: int = MQTT_BROKER_PORT,
        enable_ssl: bool = True,
    ) -> None:
        self.authorization = authorization
        self.hostname = hostname
        self.port = port
        self.vehicle_vins = []
        self.enable_ssl = enable_ssl
        self._callbacks = []
        self._operation_listeners = []
        self._listener_task = None
        self._running = False
        self._subscribed = asyncio.Event()
        self._reconnect_delay = MQTT_RECONNECT_DELAY

    async def connect(self, user_id: str, vehicle_vins: list[str]) -> None:
        """Connect to the MQTT broker and listen for messages for the given user_id and VINs."""
        _LOGGER.info("Connecting to MQTT with %s/%s", user_id, vehicle_vins)
        self.user_id = user_id
        self.vehicle_vins = vehicle_vins
        self._listener_task = asyncio.create_task(self._connect_and_listen())
        await self._subscribed.wait()

    async def disconnect(self) -> None:
        """Cancel listener task and set self_running to False, causing the listen loop to end."""
        if self._listener_task is None:
            return
        self._listener_task.cancel()
        self._listener_task = None
        self._running = False

    def subscribe(self, callback: Callable[[Event], None | Awaitable[None]]) -> None:
        """Listen for events emitted by MySkoda's MQTT broker."""
        self._callbacks.append(callback)

    def wait_for_operation(self, operation_name: OperationName) -> asyncio.Future[OperationRequest]:
        """Wait until the next operation of the specified type completes."""
        _LOGGER.debug("Waiting for operation %s complete.", operation_name)
        future: asyncio.Future[OperationRequest] = asyncio.get_event_loop().create_future()

        self._operation_listeners.append(OperationListener(operation_name, future))

        return future

    async def _connect_and_listen(self) -> None:
        """Connect to the MQTT broker and listen for messages for the given user_id and VINs.

        Reconnect loop based on https://github.com/empicano/aiomqtt/blob/main/docs/reconnection.md.

        Recreate the aiomqtt.Client on every try to get the latest authorization token.

        Passing in pre-created SSLContext (vs 'tls_params=aiomqtt.TLSParameters()') to avoid a
        blocking call in paho.mqtt.client. See https://github.com/w1ll1am23/pyeconet/pull/43.
        """
        _LOGGER.debug("Starting _connect_and_listen")
        self._running = True
        retry_count = 0  # Track the number of retries
        self._reconnect_delay = MQTT_RECONNECT_DELAY  # Initial delay for backoff
        while self._running:
            try:
                # client_id = Id + session_uuid4 + # + random_uuid4
                client_id = "Id" + str(app_uuid) + "#" + str(uuid.uuid4())
                async with aiomqtt.Client(
                    hostname=self.hostname,
                    port=self.port,
                    username="android-app",  # Explicit username from working payload
                    identifier=client_id,
                    password=await self.authorization.get_access_token(),
                    logger=_LOGGER,
                    tls_context=_SSL_CONTEXT if self.enable_ssl else None,
                    keepalive=MQTT_KEEPALIVE,
                    clean_session=True,
                ) as client:
                    _LOGGER.info("Connected to MQTT")
                    _LOGGER.debug("using MQTT client %s", client)
                    for vin in self.vehicle_vins:
                        for topic in MQTT_OPERATION_TOPICS:
                            await client.subscribe(
                                f"{self.user_id}/{vin}/operation-request/{topic}"
                            )
                        for topic in MQTT_SERVICE_EVENT_TOPICS:
                            await client.subscribe(f"{self.user_id}/{vin}/service-event/{topic}")
                        for topic in MQTT_ACCOUNT_EVENT_TOPICS:
                            await client.subscribe(f"{self.user_id}/{vin}/account-event/{topic}")
                        for topic in MQTT_VEHICLE_EVENT_TOPICS:
                            await client.subscribe(f"{self.user_id}/{vin}/vehicle-event/{topic}")

                    self._subscribed.set()
                    self._reconnect_delay = MQTT_RECONNECT_DELAY
                    retry_count = 0  # Reset retry count on successful connection
                    async for message in client.messages:
                        self._on_message(message)
            except aiomqtt.MqttError as exc:
                retry_count += 1
                _LOGGER.info(
                    "Connection lost (%s); reconnecting in %ss", exc, self._reconnect_delay
                )
                await asyncio.sleep(self._reconnect_delay)
                if (
                    retry_count > MQTT_FAST_RETRY
                    and self._reconnect_delay < MQTT_MAX_RECONNECT_DELAY
                ):  # first x retries are not exponential
                    self._reconnect_delay *= 2
                    self._reconnect_delay += uniform(0, 1)  # noqa: S311
                    self._reconnect_delay = min(self._reconnect_delay, MQTT_MAX_RECONNECT_DELAY)
                    _LOGGER.debug("Increased reconnect backoff to %s", self._reconnect_delay)

    def _on_message(self, msg: aiomqtt.Message) -> None:
        """Deserialize received MQTT message and emit Event to subscribed callbacks."""
        # Extract the topic, user id and vin from the topic's name.
        # Internally, the topic will always look like this:
        # `/{user_id}/{vin}/path/to/topic`
        topic_match = TOPIC_RE.match(str(msg.topic))
        if not topic_match:
            _LOGGER.warning("Unexpected MQTT topic encountered: %s", topic_match)
            return

        # Cast the data from binary string, ignoring empty messages.
        data = cast(str, msg.payload)
        if len(data) == 0:
            return

        self._parse_topic(topic_match, data)

    @staticmethod
    def _get_charging_event(data: str) -> ServiceEvent:
        try:
            event = ServiceEventWithChargingData.from_json(data)
        except ValueError:
            event = ServiceEvent.from_json(data)
        return event

    @staticmethod
    def _get_vehicle_ignition_status_changed_event(data: str) -> VehicleEvent:
        try:
            event = VehicleEventWithVehicleIgnitionStatusData.from_json(data)
        except ValueError:
            event = VehicleEvent.from_json(data)
        return event

    def _parse_topic(self, topic_match: re.Match[str], data: str) -> None:
        """Parse the topic and extract relevant parts."""
        [user_id, vin, event_type, topic] = topic_match.groups()
        event_type = EventType(event_type)

        _LOGGER.debug("Message (%s) received for %s on topic %s: %s", event_type, vin, topic, data)

        # Messages will contain payload as JSON.
        try:
            if event_type == EventType.OPERATION:
                self._emit(
                    EventOperation(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        operation=OperationRequest.from_json(data),
                    )
                )
            elif event_type == EventType.ACCOUNT_EVENT:
                self._emit(
                    EventAccountPrivacy(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "air-conditioning":
                self._emit(
                    EventAirConditioning(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=ServiceEvent.from_json(data),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "auxiliary-heating":
                self._emit(
                    EventAuxiliaryHeating(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=ServiceEvent.from_json(data),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "charging":
                self._emit(
                    EventCharging(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=self._get_charging_event(data),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "departure":
                self._emit(
                    EventDeparture(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=ServiceEvent.from_json(data),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "vehicle-status/access":
                self._emit(
                    EventAccess(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=ServiceEvent.from_json(data),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "vehicle-status/lights":
                self._emit(
                    EventLights(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=ServiceEvent.from_json(data),
                    )
                )
            elif event_type == EventType.SERVICE_EVENT and topic == "vehicle-status/odometer":
                self._emit(
                    EventOdometer(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=ServiceEvent.from_json(data),
                    )
                )
            elif event_type == EventType.VEHICLE_EVENT and topic == "vehicle-ignition-status":
                self._emit(
                    EventVehicleIgnitionStatus(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=self._get_vehicle_ignition_status_changed_event(data),
                    )
                )
            elif (
                event_type == EventType.VEHICLE_EVENT
                and topic == "vehicle-connection-status-update"
            ):
                self._emit(
                    EventVehicleConnectionStatusUpdate(
                        vin=vin,
                        user_id=user_id,
                        timestamp=datetime.now(tz=UTC),
                        event=VehicleEvent.from_json(data),
                    )
                )
        except Exception as exc:  # noqa: BLE001
            _LOGGER.warning("Exception parsing MQTT event: %s", exc)

    def _emit(self, event: Event) -> None:
        for callback in self._callbacks:
            result = callback(event)
            if result is not None:
                task = asyncio.create_task(cast(Any, result))
                background_tasks.add(task)
                task.add_done_callback(background_tasks.discard)

        self._handle_operation(event)

    def _handle_operation(self, event: Event) -> None:
        if event.type != EventType.OPERATION:
            return

        if event.operation.status == OperationStatus.IN_PROGRESS:
            _LOGGER.debug(
                "An operation '%s' is now in progress. Trace id: %s",
                event.operation.operation,
                event.operation.trace_id,
            )
            return

        _LOGGER.debug(
            "Operation '%s' for trace id '%s' completed.",
            event.operation.operation,
            event.operation.trace_id,
        )
        self._handle_operation_completed(event.operation)

    def _handle_operation_completed(self, operation: OperationRequest) -> None:
        listeners = self._operation_listeners
        self._operation_listeners = []
        for listener in listeners:
            if listener.operation_name != operation.operation:
                self._operation_listeners.append(listener)
                continue

            if operation.status == OperationStatus.ERROR:
                _LOGGER.error(
                    "Resolving listener for operation '%s' with error '%s'.",
                    operation.operation,
                    operation.error_code,
                )
                listener.future.set_exception(OperationFailedError(operation))
            else:
                if operation.status == OperationStatus.COMPLETED_WARNING:
                    _LOGGER.warning("Operation '%s' completed with warnings.", operation.operation)

                _LOGGER.debug("Resolving listener for operation '%s'.", operation.operation)
                listener.future.set_result(operation)
